/*
 * kanalcd.c:
 *	Text-based Kana LCD shield for π's driver.
 *	This is designed to drive the I2C interface LCD drivers
 *
 * Copyright (c) 2013 I.Kato @ BS21 Lab
 *
 ***********************************************************************
 * This file is NOT a part of wiringPi, but You have to add this to wiringPi for using.
 *
 *    kanalcd is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU Lesser General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    kanalcd is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU Lesser General Public License for more details.
 *
 *    You should have received a copy of the GNU Lesser General Public License
 *    along with wiringPi.  If not, see <http://www.gnu.org/licenses/>.
 ***********************************************************************
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdint.h>

#include <wiringPi.h>
#include <wiringPiI2C.h>

#include "kanalcd.h"


#define KANALCD_CMD_TIMEOUT  100  // タイムアウト時間:100ms (SD1602H は画面クリアに約1.52ms＋α かかるので、それよりも大きい値にしてください)


#define KANALCD_I2C_STOP   0   // stop bit  に相当するアドレス
#define KANALCD_I2C_DATA1  1   // data 1    に相当するアドレス
#define KANALCD_I2C_DATA0  2   // data 0    に相当するアドレス
#define KANALCD_I2C_START  3   // start bit に相当するアドレス


// コマンド・ビット
#define KANALCD_CMD_INIT            0b00000001  // 初期化コマンド
#define KANALCD_CMD_RET_HOME        0b00000010  // HomeReturn コマンド
#define KANALCD_CMD_CLEAR           0b00000011  // 画面クリアコマンド
#define KANALCD_CMD_BLINK           0b0001000   // カーソル点滅 ON/OFF コマンド
#define KANALCD_CMD_CURSOR          0b0001001   // カーソル表示 ON/OFF コマンド
#define KANALCD_CMD_DISPLAY         0b0001010   // 画面表示 ON/OFF コマンド
#define KANALCD_CMD_ENTRY_MODE_SET  0b000110    // EntryMode 設定コマンド
#define KANALCD_CMD_SHIFT           0b000111    // Shift コマンド
#define KANALCD_CMD_LOCATE          0b001       // Locate コマンド
#define KANALCD_CMD_USER_DEF_FONT   0b01000     // ユーザー定義フォントコマンド
#define KANALCD_CMD_DIGITAL_WRITE   0b0110      // デジタル出力コマンド
#define KANALCD_CMD_PIN_MODE        0b01110     // ピンモード設定コマンド
#define KANALCD_CMD_PWM_WRITE       0b01111     // PWM 出力コマンド
#define KANALCD_CMD_TEXT            0b1         // Text コマンド


#define KANALCD_PORT_BACKLIGHT  7  // SD1602H の Backlight ON/OFF用


// Raspberry Pi に使用されている Broadcom BCM2835 の i2c には、clock stretching のバグがあります。
//     ⇒ 詳細は、http://www.advamation.com/knowhow/raspberrypi/rpi-i2c-bug.html を参照（英文です）

// i2c 通信は、例えば、マスターからスレーブへの書き込み(SlaveRecieve)の場合、
//  (1) i2c バスは解放されている
//  (2) マスターが start condition を発行し、i2c バスの制御権を掌握する
//  (3) マスターが、通信したいスレーブのアドレスと R/W bit を送信する
//  (4) 該当のスレーブが ACK か NACK を返送する
//  (5) NACK だった場合、(9) に行く
//  (6) ACK だった場合、マスターはデータ 8bit を送信する
//  (7) 該当のスレーブは送信されたデータ 8bit を受信し、ACK か NACK を返送する
//  (8) ACK なら、(6)～(7) を繰り返す
//  (9) NACK だった時や、送るべきデータがなくなった時は、マスターは stop condition を発行する
//  (10) i2c バスは解放される
// という感じで i2c 通信が行われますが、(4)や(7)の時、スレーブが ACK か NACK を返送する時
// スレーブ側の要請で clock stretching が発生する場合が有ります。
//（100％発生ではなく、低速応答のスレーブでのみ発生し、十分に高速なスレーブなら発生しない）

// マスターとなるべき Broadcom BCM2835 は、この clock stretching が起きた時の処理にバグがあるため、
// clock stretching が発生する i2c デバイスとは、上手く i2c 通信できない場合があります。
// 具体的には、上記(6)の時の、データ 8bit 送信での１ビット目のクロック幅が、仕様を満たさない状態になります。
// この為、スレーブ側では、数バイトは正常に受信するのに、次の数バイトは１ビットずつズレてしまう現象や、
// スレーブ側が 8bit 分を受信する前に stop condition が発行されて、スレーブ側が対応に困る、という現象が発生するようです。
// このバグは BCM2835 のハードウエア・レベルのバグですので、ソフト的に回避する方法は無いようです。

// カナ LCD シールド(π用)は、i2c 通信のスレーブとして AVR を使用していますが、
// やはり clock stretching が発生するらしく(オシロで波形を見たわけでないので断言できない)、
// 当初の i2c 通信方法では、盛大にコケてました。
// そこで対策として採ったのが、スレーブアドレスそのものでデータ送信する方法です。
// BCM2835 のバグは、上記(6)以降で問題になるので、それよりも前のスレーブアドレス部分だけは
// 安定して送信できているだろう、と考えたからです。
// 具体的には４つのスレーブアドレスを確保して、１つをパケット開始に、１つをパケット終了に、
// 残りの２つをデータ0/1用に使用する事にしました。

// ただし、この方法でも、マスターからスレーブへの送信はできますが、スレーブからマスターへの送信はできません。
// スレーブから送信されるデータ 8bit も clock stretching バグの影響を受けるからです。
// 唯一、スレーブアドレス直後に返送する ACK/NACK を使って 1bit だけスレーブのデータを送信する方法は可能です。
// カナ LCD シールド(π用)では、コマンド受付け可の時は ACK を、コマンド処理中で新規コマンド受付け不可の時は NACK を返す様にしています。
// (スレーブアドレスが４つあるので、各々の ACK/NACK に別々の意味を持たせれば 4bit のデータ送信も可能になりますが、そこまではサピンしませんでした)


// カナ LCD シールド(π用)管理構造体
struct kanalcdDataStruct
{
	int     fdStop;
	int     fdData1;
	int     fdData0;
	int     fdStart;
};

struct kanalcdDataStruct kanalcds [MAX_KANALCDS];  // 最大４台のカナLCDシールド(π用)を管理する：正直、４台も接続する事ないと思いますが、wiringPiのlcd.hに合わせておきます



// カナ LCD シールド(π用)がコマンド受付け可能になるまで待つ
// 戻り値が KANALCD_ERR_NONE なら、コマンド受付け可能状態
// 注意：シールドの制御IC (ATmega48A) の clock は約 8MHz なので、それなりに待ち時間を入れる必要があります。
static int waitWhileBusy (int fd)
{
	int res;
	unsigned int t = micros(); // wiringPiSetup() からの経過時間(μ秒)

	delayMicroseconds (10); // slave側の処理時間の確保

	while (1) { // 永久ループ
		res = wiringPiI2CWriteQuick (fd); // stop bit を送ってみる
		if (res == KANALCD_ERR_NONE) {
			// ACK が返ってきた
			delayMicroseconds (10); // slave側の処理時間の確保
			return KANALCD_ERR_NONE; // コマンド受付け可能状態
		}
//		printf("waitWhileBusy(): res = %02x\n", res);
		// 「まだだよ」返答：カナ LCD シールドは以前のコマンドを処理中(SLA+W に対して NACK が返ってきた)か、そもそも、TWI 通信自体ができていない
		if ((micros() - t) > KANALCD_CMD_TIMEOUT * 1000) { // 所定の時間が経過した (おそらく TWI 通信できていない)
			break; // 強制ループアウトさせる
		}
		delayMicroseconds (50); // 50μ秒待つ
	}
//	printf("waitWhileBusy(): TimeOut.\n");
	return KANALCD_ERR_TIME_OUT; // タイムアウトした
}


// カナ LCD シールド(π用)に i2c 通信で bit を送信する
static int sendI2CBit (int fd)
{
	int res = wiringPiI2CWriteQuick (fd); // bit を送る

	delayMicroseconds (10); // slave側の処理時間の確保

	return res;
}

// len ビット分の bit 列を送信する
static int sendI2CBits (int fdIdx, unsigned long bits, int len)
{
	int res = -1;
	while (len) {
		--len;
		res = sendI2CBit ((bits & (1uL << len)) ? kanalcds[fdIdx].fdData1 : kanalcds[fdIdx].fdData0);
	}
	return res;
}


/*
 *********************************************************************************
 * User Callable code below here
 *********************************************************************************
 */



/*
 * kanalcdInit:
 *	Take a lot of parameters and initialise the LCD, and return a handle to
 *	that LCD, or -1 if any error.
   シールド上の DIPSW3 で設定している i2c アドレス(7bit format)を指定して下さい。
   建前上、file descriptor を返す、事になってますが、実際は kanalcds 配列のインデックス値(0～3)です。
   エラーの時は -1 を返します。
   正常に実行された場合は、画面クリア、カーソル位置(0,0)、LCD 表示、カーソル非点滅、カーソル非表示、
   バックライト・ソフトOFF、汎用出力ピン All デジタル出力OFF になります。
 *********************************************************************************
 */

int kanalcdInit (int i2c_addr)
{
	static int initialised = 0;

	int i, res;
	int fdIdx = -1;

	if (initialised == 0) { // 初期化されていない
		initialised = 1;
		for (i = 0 ; i < MAX_KANALCDS ; ++i) {
			kanalcds[i].fdStop = -1;
		}
	}

	// 空いているカナ LCD シールド(π用)管理構造体を探します。
	for (i = 0 ; i < MAX_KANALCDS ; ++i) {
		if (kanalcds[i].fdStop == -1) {
			fdIdx = i;
			break;
		}
	}

	if (fdIdx == -1) {
		// 既に4台確保している時はエラー
		return KANALCD_ERR_FULL;
	}

	// 初期化
	// i2c アドレス連続４つを確保します。
	kanalcds[fdIdx].fdStop = wiringPiI2CSetup (i2c_addr + KANALCD_I2C_STOP);
	if (kanalcds[fdIdx].fdStop == -1) {
		return KANALCD_ERR_NO_DEV;
	}
	kanalcds[fdIdx].fdData1 = wiringPiI2CSetup (i2c_addr + KANALCD_I2C_DATA1);
	if (kanalcds[fdIdx].fdData1 == -1) {
		return KANALCD_ERR_NO_DEV;
	}
	kanalcds[fdIdx].fdData0 = wiringPiI2CSetup (i2c_addr + KANALCD_I2C_DATA0);
	if (kanalcds[fdIdx].fdData0 == -1) {
		return KANALCD_ERR_NO_DEV;
	}
	kanalcds[fdIdx].fdStart = wiringPiI2CSetup (i2c_addr + KANALCD_I2C_START);
	if (kanalcds[fdIdx].fdStart == -1) {
		return KANALCD_ERR_NO_DEV;
	}

	res = waitWhileBusy (kanalcds[fdIdx].fdStop);

	if (res == KANALCD_ERR_NONE) {

		// カナ LCD シールドへ初期化コマンド(0000 0001)を送信
		sendI2CBit (kanalcds[fdIdx].fdStart);

		sendI2CBits (fdIdx, KANALCD_CMD_INIT, 8);

		sendI2CBit (kanalcds[fdIdx].fdStop);

		// 続けて、画面クリアコマンドも実行します
		kanalcdClear (fdIdx);
	}

//	printf("kanalcdInit(): done.\n");

	return fdIdx;
}


/*
 * kanalcdHome:
 *	Home the cursor.
    カーソルを (0,0) の位置にリターンさせます。処理に約1.52ms＋α かかります。
    (＋α分ですが、カナLCDシールドで使用している制御IC が内蔵RC発振なので ±10%前後の誤差が有るため、確実に処理が完了した事を保証させる為に約0.25ms を追加してあります)
    このコマンドそのものは、カナLCDシールドに指示したら即リターンしますので、この次の命令で最大 1.52ms＋α 待たされる事になります。
    なので、kanalcdHome 関数と次の命令の間に、1.52ms 以上かかる別の処理をやらせると良いでしょう。
 *********************************************************************************
 */

int kanalcdHome (int fdIdx)
{
	int res = waitWhileBusy (kanalcds[fdIdx].fdStop);

	if (res == KANALCD_ERR_NONE) {

		// Return Home コマンド (0000 0010)
		sendI2CBit (kanalcds[fdIdx].fdStart);

		sendI2CBits (fdIdx, KANALCD_CMD_RET_HOME, 8);

		sendI2CBit (kanalcds[fdIdx].fdStop);
	}

	return res;
}


/*
 * kanalcdClear:
 *	clear the screen.
    画面をクリアします。クリア処理に約1.52ms＋α かかります。
    (＋α分ですが、カナLCDシールドで使用している制御IC が内蔵RC発振なので ±10%前後の誤差が有るため、確実に処理が完了した事を保証させる為に約0.25ms を追加してあります)
    このコマンドそのものは、カナLCDシールドに指示したら即リターンしますので、この次の命令で最大 1.52ms＋α 待たされる事になります。
    なので、kanalcdClear 関数と次の命令の間に、1.52ms 以上かかる別の処理をやらせると良いでしょう。
 *********************************************************************************
 */

int kanalcdClear (int fdIdx)
{
	int res = waitWhileBusy (kanalcds[fdIdx].fdStop);

	if (res == KANALCD_ERR_NONE) {

		// 画面クリアコマンド (0000 0011)
		sendI2CBit (kanalcds[fdIdx].fdStart);

		sendI2CBits (fdIdx, KANALCD_CMD_CLEAR, 8);

		sendI2CBit (kanalcds[fdIdx].fdStop);
	}

	return res;
}


/*
 * kanalcdDisplay:
 *	Turn the display on/off
    液晶画面の表示を ON/OFF します。
    この命令では backlight の方は無変化です。
    kanalcdInit 関数を呼び出した直後は ON になっています。
 *********************************************************************************
 */

int kanalcdDisplay (int fdIdx, int state)
{
	int res = waitWhileBusy (kanalcds[fdIdx].fdStop);

	if (res == KANALCD_ERR_NONE) {

		// 画面表示 ON/OFF コマンド (0001 010s)
		sendI2CBit (kanalcds[fdIdx].fdStart);

		sendI2CBits (fdIdx, KANALCD_CMD_DISPLAY, 7);

		sendI2CBit (state ? kanalcds[fdIdx].fdData1 : kanalcds[fdIdx].fdData0); // ON/OFF bit

		sendI2CBit (kanalcds[fdIdx].fdStop);
	}

	return res;
}

/*
 * kanalcdCursor: 
 *	Turn the cursor on/off
    カーソル表示を ON/OFF します。
    kanalcdInit 関数を呼び出した直後は OFF になっています。
 *********************************************************************************
 */

int kanalcdCursor (int fdIdx, int state)
{
	int res = waitWhileBusy (kanalcds[fdIdx].fdStop);

	if (res == KANALCD_ERR_NONE) {

		// カーソル表示 ON/OFF コマンド (0001 001s)
		sendI2CBit (kanalcds[fdIdx].fdStart);

		sendI2CBits (fdIdx, KANALCD_CMD_CURSOR, 7);

		sendI2CBit (state ? kanalcds[fdIdx].fdData1 : kanalcds[fdIdx].fdData0); // ON/OFF bit

		sendI2CBit (kanalcds[fdIdx].fdStop);
	}

	return res;
}

/*
 * kanalcdCursorBlink:
 *	Turn the cursor blinking on/off
    カーソルの点滅を ON/OFF します。
    kanalcdInit 関数を呼び出した直後は OFF になっています。
 *********************************************************************************
 */

int kanalcdCursorBlink (int fdIdx, int state)
{
	int res = waitWhileBusy (kanalcds[fdIdx].fdStop);

	if (res == KANALCD_ERR_NONE) {

		// カーソル点滅 ON/OFF コマンド (0001 000s)
		sendI2CBit (kanalcds[fdIdx].fdStart);

		sendI2CBits (fdIdx, KANALCD_CMD_BLINK, 7);

		sendI2CBit (state ? kanalcds[fdIdx].fdData1 : kanalcds[fdIdx].fdData0); // ON/OFF bit

		sendI2CBit (kanalcds[fdIdx].fdStop);
	}

	return res;
}


/*
 * kanalcdEntryModeSet:
    Entry Mode を設定します。
    機能の詳細は SD1602H, HD44780 のデータシートを参照して下さい。
    kanalcdInit 関数を呼び出した直後は I/D=ON(INC), S=OFF になっています。
 *********************************************************************************
 */

int kanalcdEntryModeSet (int fdIdx, unsigned int ems)
{
	int res = waitWhileBusy (kanalcds[fdIdx].fdStop);

	if (res == KANALCD_ERR_NONE) {

		// Entry Mode Set コマンド (0001 10xx)
		sendI2CBit (kanalcds[fdIdx].fdStart);

		sendI2CBits (fdIdx, KANALCD_CMD_ENTRY_MODE_SET, 6);

		sendI2CBit ((ems & 0x02) ? kanalcds[fdIdx].fdData1 : kanalcds[fdIdx].fdData0); // I/D bit
		sendI2CBit ((ems & 0x01) ? kanalcds[fdIdx].fdData1 : kanalcds[fdIdx].fdData0); // S bit

		sendI2CBit (kanalcds[fdIdx].fdStop);
	}

	return res;
}


/*
 * kanalcdShift:
    Cursor/Display Shift を実行します。
    機能の詳細は SD1602H, HD44780 のデータシートを参照して下さい。
 *********************************************************************************
 */

int kanalcdShift (int fdIdx, unsigned int shift)
{
	int res = waitWhileBusy (kanalcds[fdIdx].fdStop);

	if (res == KANALCD_ERR_NONE) {

		// Cursor/Display Shift コマンド (0001 11xx)
		sendI2CBit (kanalcds[fdIdx].fdStart);

		sendI2CBits (fdIdx, KANALCD_CMD_SHIFT, 6);

		sendI2CBit ((shift & 0x02) ? kanalcds[fdIdx].fdData1 : kanalcds[fdIdx].fdData0); // S/C bit
		sendI2CBit ((shift & 0x01) ? kanalcds[fdIdx].fdData1 : kanalcds[fdIdx].fdData0); // R/L bit

		sendI2CBit (kanalcds[fdIdx].fdStop);
	}

	return res;
}


/*
 * kanalcdPosition:
 *	Update the position of the cursor on the display.
 *	Ignore invalid locations.
    カーソルの位置を指定します。指定できるのは
    col は 0～39 の範囲 (ただし表示されるのは、この内の16桁のみ)
    row は 0～1 の範囲、です。
 *********************************************************************************
 */

int kanalcdPosition (int fdIdx, unsigned int col, unsigned int row)
{
	if (col >= KANALCD_MAX_COLS) { // 範囲外
		return KANALCD_ERR_OUT_OF_RANGE; // 引数が不正
	}
	if (row >= KANALCD_MAX_ROWS) { // 範囲外
		return KANALCD_ERR_OUT_OF_RANGE; // 引数が不正
	}

	int res = waitWhileBusy (kanalcds[fdIdx].fdStop);

	if (res == KANALCD_ERR_NONE) {

		// カーソル Locate コマンド (001 y xxxxxx) : 10bit長
		sendI2CBit (kanalcds[fdIdx].fdStart);

		sendI2CBits (fdIdx, KANALCD_CMD_LOCATE, 3);

		sendI2CBits (fdIdx, row, 1);  // y座標指定が先

		sendI2CBits (fdIdx, col, 6);  // x座標指定は後

		sendI2CBit (kanalcds[fdIdx].fdStop);
	}

	return res;
}


/*
 * kanalcdCharDef:
 *	Defines a new character in the CGRAM
    カナ LCD シールド(π用)にユーザー定義文字フォント(5x8dots)を設定します。
    index=0～7 の範囲。
    設定したユーザー定義文字は、全角の'１'～'８'か'①'～'⑧'で表示できます
 *********************************************************************************
 */

int kanalcdCharDef (int fdIdx, unsigned int index, unsigned char font_data[8])
{
	int i, res;

	if (index >= KANALCD_MAX_CHARDEFS) { // 最大８個まで
		return KANALCD_ERR_OUT_OF_RANGE;
	}

	res = waitWhileBusy (kanalcds[fdIdx].fdStop);

	if (res == KANALCD_ERR_NONE) {

		// CHARDEF コマンド (0100 0nnn bbbbb .... bbbbb) : 48bit長
		sendI2CBit (kanalcds[fdIdx].fdStart);

		sendI2CBits (fdIdx, KANALCD_CMD_USER_DEF_FONT, 5);

		sendI2CBits (fdIdx, index, 3);  // 0～7 の範囲

		for (i = 0; i < 8; i++) { // 縦 8 dots
			// 横 5 dots : 下位 5bit のみ有効。上位 3bit は無視されます。
			sendI2CBits (fdIdx, font_data[i], 5);
		}

		sendI2CBit (kanalcds[fdIdx].fdStop);
	}

	return res;
}


/*
 * kanalcdPutchar:
 *	Send a data byte to be displayed on the display. We implement a very
 *	simple terminal here - with line wrapping, but no scrolling. Yet.
    カナ LCD シールド(π用)に文字コード１バイトを送信します。
    カナ LCD シールド(π用)の方でマルチバイト処理しますので、１バイトずつ転送するだけで OK です
 *********************************************************************************
 */

int kanalcdPutchar (int fdIdx, char ch)
{
	int res = waitWhileBusy (kanalcds[fdIdx].fdStop);

	if (res == KANALCD_ERR_NONE) {

//		printf("kanalcdPutchar(): ch = %02x\n", ch);

		// テキスト転送コマンド (1 cccccccc) : 9bit長
		sendI2CBit (kanalcds[fdIdx].fdStart);

		sendI2CBit (kanalcds[fdIdx].fdData1);  // Textコマンド

		sendI2CBits (fdIdx, ch, 8);

		sendI2CBit (kanalcds[fdIdx].fdStop);
	}

	return res;
}


/*
 * kanalcdPuts:
 *	Send a string to be displayed on the display
    文字列をカナLCDシールドに送信、表示させます。
    自動的に改行しないので、改行させたい場合は'\n'を指定する必要があります。
    正常表示できた場合は、バイト数を返します。
 *********************************************************************************
 */

int kanalcdPuts (int fdIdx, const char * str)
{
	char ch;
	int res, len = 0;

	if (str == NULL) {
		return KANALCD_ERR_INVALID_PARAM;
	}

	while (ch = *str++) {
		res = kanalcdPutchar (fdIdx, ch) ;
		if (res != KANALCD_ERR_NONE) {
			return res;
		}
		len++;
	}
	return len;
}


/*
 * kanalcdPrintf:
 *	Printf to an LCD display
    printf の書式に従って、文字列をカナLCDシールドに送信、表示させます。
    正常表示できた場合は、バイト数を返します。
 *********************************************************************************
 */

int kanalcdPrintf (int fdIdx, const char * fmt, ...)
{
	va_list argp ;
	char buffer [1024] ;

	if (fmt == NULL) {
		return KANALCD_ERR_INVALID_PARAM;
	}

	va_start  (argp, fmt) ;
	vsnprintf (buffer, 1023, fmt, argp) ;
	va_end    (argp) ;

	return kanalcdPuts (fdIdx, buffer) ;
}



// シールド上の DIPSW3 で、バックライト・ソフトON/OFF モードにしている場合のみ有効。
// バックライト強制ON モードにしている場合は、以下のコマンドでの ON/OFF に関係無く、バックライトON されます。
// kanalcdInit 関数を呼び出した直後は、バックライト・ソフトOFF になります。
int kanalcdBacklight (int fdIdx, int state)
{
	int res = waitWhileBusy (kanalcds[fdIdx].fdStop);

	if (res == KANALCD_ERR_NONE) {

		// デジタル出力コマンド (0110 111s)
		sendI2CBit (kanalcds[fdIdx].fdStart);

		sendI2CBits (fdIdx, KANALCD_CMD_DIGITAL_WRITE, 4);

		sendI2CBits (fdIdx, KANALCD_PORT_BACKLIGHT, 3);

		sendI2CBit (state ? kanalcds[fdIdx].fdData1 : kanalcds[fdIdx].fdData0);  // ON/OFF bit

		sendI2CBit (kanalcds[fdIdx].fdStop);
	}

	return res;
}



// 汎用出力関係
// 注意：デジタル出力も PWM 出力も、共に +5V 出力になります。+3.3V ではありませんので、注意して下さい。

// ※ 当初、PWM 機能が 2ch しかない ATtiny48 を使用する予定だった為、PWM0 と PWM1 の２つのみになる予定でした（OUT0 はデジタル出力のみだった）。
//    しかし、諸事情で ATmega48A に変更になった時に、OUT0 を PWM2 に名称変更するべきだったのですが、思いっきり見落として、基板を作ってしまいました。
//    分かりやすさから言えば、PWM0/PWM1/PWM2/OUT3 という風にした方が絶対に良かったのですが、基板作り直しにお金が掛かるので、申し訳有りませんが、'現状優先'とさせて下さい。


// 指定の汎用出力ピン(PWM0/PWM1/OUT0)を OUTPUTモードか PWM_OUTPUT モードに設定します。
// kanalcdInit 関数を呼び出した直後は、３点とも OUTPUTモードになります。
int kanalcdPinMode (int fdIdx, unsigned int pin, int mode)
{
	if (pin >= KANALCD_MAX_PINS) {
		return KANALCD_ERR_OUT_OF_RANGE;
	}

	int res = waitWhileBusy (kanalcds[fdIdx].fdStop);

	if (res == KANALCD_ERR_NONE) {

		// ピンモード設定コマンド (0111 0ppm)
		sendI2CBit (kanalcds[fdIdx].fdStart);

		sendI2CBits (fdIdx, KANALCD_CMD_PIN_MODE, 5);

		sendI2CBits (fdIdx, pin, 2);  // 0～2のみ。3は何も影響しない。

		sendI2CBit ((mode == PWM_OUTPUT) ? kanalcds[fdIdx].fdData1 : kanalcds[fdIdx].fdData0);  // mode bit

		sendI2CBit (kanalcds[fdIdx].fdStop);
	}

	return res;
}


// 指定の汎用出力ピン(PWM0/PWM1/OUT0/OUT1)をデジタル ON/OFF させます。
// kanalcdInit 関数を呼び出した直後は、４点とも OFF になります。
// 注意：+5V 仕様です。+3.3V ではありません。
int kanalcdDigitalWrite (int fdIdx, unsigned int pin, int state)
{
	if (pin >= KANALCD_MAX_PINS) {
		return KANALCD_ERR_OUT_OF_RANGE;
	}

	int res = waitWhileBusy (kanalcds[fdIdx].fdStop);

	if (res == KANALCD_ERR_NONE) {

		// デジタル出力コマンド (0110 ppps)
		sendI2CBit (kanalcds[fdIdx].fdStart);

		sendI2CBits (fdIdx, KANALCD_CMD_DIGITAL_WRITE, 4);

		sendI2CBits (fdIdx, pin, 3);  // 0～3のみ。4～6は何も影響しない。7はBacklightで使用。

		sendI2CBit (state ? kanalcds[fdIdx].fdData1 : kanalcds[fdIdx].fdData0);  // ON/OFF bit

		sendI2CBit (kanalcds[fdIdx].fdStop);
	}

	return res;
}


// 指定の汎用出力ピン(PWM0/PWM1/OUT0)から PWM 出力させます。
// kanalcdPinMode 関数で PWM_OUTPUTモードに設定している必要があります。
// 指定できる PWM 値は 0～255 の範囲です。
// 注意：+5V 仕様です。+3.3V ではありません。
int kanalcdPwmWrite (int fdIdx, unsigned int pin, unsigned int val)
{
	if (pin >= KANALCD_MAX_PWMS) {
		return KANALCD_ERR_OUT_OF_RANGE;
	}

	int res = waitWhileBusy (kanalcds[fdIdx].fdStop);

	if (res == KANALCD_ERR_NONE) {

		// PWM出力コマンド (0111 1pp vvvvvvvv) : 15bit長
		sendI2CBit (kanalcds[fdIdx].fdStart);

		sendI2CBits (fdIdx, KANALCD_CMD_PWM_WRITE, 5);

		sendI2CBits (fdIdx, pin, 2);  // 0～2のみ。3は何も影響しない。

		sendI2CBits (fdIdx, val, 8);  // PWM 値：0～255

		sendI2CBit (kanalcds[fdIdx].fdStop);
	}

	return res;
}

