/*
 * kanalcd.h:
 *	Text-based Kana LCD shield for π's driver.
 *	This is designed to drive the I2C interface LCD drivers
 *
 * Copyright (c) 2013 I.Kato @ BS21 Lab
 *
 ***********************************************************************
 * This file is NOT a part of wiringPi, but You have to add this to wiringPi for using.
 *
 *    kanalcd is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU Lesser General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    kanalcd is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU Lesser General Public License for more details.
 *
 *    You should have received a copy of the GNU Lesser General Public License
 *    along with wiringPi.  If not, see <http://www.gnu.org/licenses/>.
 ***********************************************************************
 */


// カナ LCD シールド(π用)

// RaspberryPi から i2c通信で SD1602H を制御します。
// 使用する i2c アドレスは、0x0C～0x0F, 0x1C～0x1F, 0x2C～0x2F, 0x3C～0x3F の４つから選択できます。
// 必ず、連続４アドレス（例えば、0x3C,0x3D,0x3E,0x3F）をごっそりと占有しますので、注意して下さい。

// 表示できる文字は SD1602H (制御IC:HD44780) で定義されている文字と、ユーザー定義文字８文字のみです。
// 使用できる符号化方式は UTF-8 のみです。他の符号化方式には変更できません。
// UTF-8コードから HD44780コードへの変換は、シールド側で行われますので、特に気にする必要は有りません。
// ASCII の ~ (チルダ,0x7E) は HD44780 に含まれていないので表示できません。代わりに'-1乗'記号が表示されます。
// ユーザー定義文字は、全角文字の'１'～'８'か'①'～'⑧'を指定すれば、表示させる事ができます。
// 本シールド電源 ON 直後は、ユーザー定義文字は、日,月,火,水,木,金,土,ハート記号 が設定されています。
// もちろん、kanalcdCharDef 関数を使って、ユーザーが自由に変更できます。

// SD1602H 表示機能以外に、汎用出力ピン４点があります。ただし +5V 出力ですので、注意が必要です。
// 汎用出力ピン４点の内、３点は PWM 出力(8bit長)も可能です。こちらも +5V 出力になります。



// 最大４台のカナ LCD シールド(π用)を管理します。
// 正直、４台も接続する事ないと思いますが、wiringPi の lcd.h に合わせておきます。
// 当然ですが、各々の i2c アドレスは重複しないように設定する必要が有ります。
#define MAX_KANALCDS   4


// カナ LCD シールド(π用)の最大桁数・行数
#define KANALCD_MAX_COLS    40  // 最大桁数：実際に表示されるのは40桁中の16桁のみです。残り24桁はシフトさせなければ見えません。
#define KANALCD_MAX_ROWS     2  // 最大行数

#define KANALCD_MAX_CHARDEFS  8  // ユーザー定義文字の最大値


#define KANALCD_STATE_OFF   0
#define KANALCD_STATE_ON    1


#define KANALCD_EMS_AC_INC     0b10  // default
#define KANALCD_EMS_AC_DEC     0b00
#define KANALCD_EMS_SHIFT_OFF  0b00  // default
#define KANALCD_EMS_SHIFT_ON   0b01


#define KANALCD_SHIFT_CURSOR_LEFT    0b00  // カーソルを左に１つシフトさせる
#define KANALCD_SHIFT_CURSOR_RIGHT   0b01  // カーソルを右に１つシフトさせる
#define KANALCD_SHIFT_DISPLAY_LEFT   0b10  // 画面を左に１つシフトさせる
#define KANALCD_SHIFT_DISPLAY_RIGHT  0b11  // 画面を右に１つシフトさせる



// 汎用出力ピン：４点。この内３点は PWM 出力も可能。
#define KANALCD_PIN_PWM0   0  // シルク'PWM0' ※デジタルON/OFF出力も可能
#define KANALCD_PIN_PWM1   1  // シルク'PWM1' ※デジタルON/OFF出力も可能
#define KANALCD_PIN_OUT0   2  // シルク'OUT0' ※PWM出力も可能
#define KANALCD_PIN_OUT1   3  // シルク'OUT1' ※デジタルON/OFF出力のみ。PWM出力はできない

#define KANALCD_MAX_PINS   4  // デジタル出力できる最大ピン数
#define KANALCD_MAX_PWMS   3  // PWM出力できる最大ピン数


// カナ LCD ライブラリ(π用)のエラーコード
#define KANALCD_ERR_NONE             0    // 正常
#define KANALCD_ERR_TIME_OUT       (-15)  // カナ LCD シールドとの通信がタイムアウトした
#define KANALCD_ERR_NO_DEV         (-12)  // 指定した i2c アドレスのカナ LCD シールドは無かった
#define KANALCD_ERR_FULL           (-11)  // ４台以上を確保しようとした
#define KANALCD_ERR_INSUF_MEM      (-10)  // メモリが足らない
#define KANALCD_ERR_OUT_OF_RANGE   (-6)   // 指定できる範囲から逸脱している
#define KANALCD_ERR_INVALID_PARAM  (-5)   // 引数が不正
#define KANALCD_ERR_NOT_INIT       (-3)   // 初期化されていない
#define KANALCD_ERR_UNKNOWN        (-1)   // 原因不明のエラーが発生した


#ifdef __cplusplus
extern "C" {
#endif


// カナ LCD シールド(π用)を初期化する。
// シールド上の DIPSW3 で設定している i2c アドレス(7bit format)を指定して下さい。
// 正常に実行された場合は、画面クリア、カーソル位置(0,0)、LCD 表示、カーソル非点滅、カーソル非表示、バックライト・ソフトOFF、汎用出力ピン All OFF になります
extern int kanalcdInit (int i2c_addr);  // file descriptor を返します。エラーの時は -1


// wiringPi の lcd.h 互換関数

extern int kanalcdHome        (int fd);  // 画面はそのままで、カーソル位置だけ (0,0) にする ※処理に約 1.52ms＋α かかります
extern int kanalcdClear       (int fd);  // 画面をクリアし、カーソル位置を (0,0) にする ※処理に約 1.52ms＋α かかります
extern int kanalcdDisplay     (int fd, int state);  // LCD を表示させる/させない
extern int kanalcdCursor      (int fd, int state);  // カーソルを表示させる/させない
extern int kanalcdCursorBlink (int fd, int state);  // カーソル位置の文字を点滅させる/させない
extern int kanalcdPosition    (int fd, unsigned int col, unsigned int row);  // カーソル位置を設定: col=0～39, row=0～1 の範囲で指定
extern int kanalcdCharDef     (int fd, unsigned int index, unsigned char font_data[8]);  // ユーザー定義文字フォント(5x8dots)を設定: index=0～7 の範囲。
extern int kanalcdPutchar     (int fd, char ch);  // 文字コードを１バイト転送します
extern int kanalcdPuts        (int fd, const char * str);  // 文字列を転送する。自動改行はしないので'\n'を指定する必要あり。
extern int kanalcdPrintf      (int fd, const char * fmt, ...);  // printf


// カナ LCD シールド(π用)専用の関数

// SD1602H 追加コマンド
extern int kanalcdEntryModeSet (int fd, unsigned int ems);    // Entry Mode を設定します。
extern int kanalcdShift        (int fd, unsigned int shift);  // Cursor/Display Shift を実行します。

// シールド上の DIPSW3 で、バックライト・ソフトON/OFF モードにしている場合のみ有効。
// バックライト強制ON モードにしている場合は、以下のコマンドに関係無く、バックライトON される
extern int kanalcdBacklight    (int fd, int state); // LCD バックライトを ON/OFF させる

// 汎用出力ピン
extern int kanalcdPinMode      (int fd, unsigned int pin, int mode);          // 指定の汎用出力ピン(PWM0/PWM1/OUT0)を OUTPUTモードか PWM_OUTPUT モードに設定する
extern int kanalcdDigitalWrite (int fd, unsigned int pin, int state);         // 指定の汎用出力ピン(PWM0/PWM1/OUT0/OUT1)をデジタル ON/OFF させる
extern int kanalcdPwmWrite     (int fd, unsigned int pin, unsigned int val);  // 指定の汎用出力ピン(PWM0/PWM1/OUT0)から PWM 出力させる。PWM_OUTPUTモードになっている必要あり。指定できる値は 0～255 の範囲。


#ifdef __cplusplus
}
#endif

