/*
 * knjlcd.c:
 *	Text-based Kanji LCD shield for π's driver.
 *	This is designed to drive the I2C interface LCD drivers
 *
 * Copyright (c) 2013 I.Kato @ BS21 Lab
 *
 ***********************************************************************
 * This file is NOT a part of wiringPi, but You have to add this to wiringPi for using.
 *
 *    knjlcd is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU Lesser General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    knjlcd is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU Lesser General Public License for more details.
 *
 *    You should have received a copy of the GNU Lesser General Public License
 *    along with wiringPi.  If not, see <http://www.gnu.org/licenses/>.
 ***********************************************************************
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdint.h>

#include <wiringPi.h>
#include <wiringPiI2C.h>

#include "knjlcd.h"


#define KNJLCD_CMD_TIMEOUT  100  // タイムアウト時間:100ms (画面クリアに約 60ms かかるので、それよりも大きい値にしてください)


#define KNJLCD_I2C_STOP   0   // stop bit  に相当するアドレス
#define KNJLCD_I2C_DATA1  1   // data 1    に相当するアドレス
#define KNJLCD_I2C_DATA0  2   // data 0    に相当するアドレス
#define KNJLCD_I2C_START  3   // start bit に相当するアドレス


// コマンド・ビット
#define KNJLCD_CMD_INIT              0b00000001  // 初期化コマンド
#define KNJLCD_CMD_RET_HOME          0b00000010  // HomeReturn コマンド
#define KNJLCD_CMD_CLEAR             0b00000011  // 画面クリアコマンド
//#define KNJLCD_CMD_ENCODING_SET    0b00001     // 無し：Encoding 設定コマンド ※UTF-8専用なので不要
#define KNJLCD_CMD_BLINK             0b0001000   // カーソル点滅 ON/OFF コマンド
#define KNJLCD_CMD_CURSOR            0b0001001   // カーソル表示 ON/OFF コマンド
#define KNJLCD_CMD_DISPLAY           0b0001010   // 画面表示 ON/OFF コマンド
//#define KNJLCD_CMD_ENTRY_MODE_SET  0b000110    // 無し：EntryMode 設定コマンド
//#define KNJLCD_CMD_SHIFT           0b000111    // 無し：Shift コマンド
#define KNJLCD_CMD_LOCATE            0b001       // Locate コマンド
//#define KNJLCD_CMD_USER_DEF_FONT   0b01000     // 無し：ユーザー定義フォントコマンド
#define KNJLCD_CMD_GRAPH_PSET        0b0101001   // Pset コマンド
#define KNJLCD_CMD_GRAPH_LINE        0b0101010   // Line コマンド
#define KNJLCD_CMD_GRAPH_RECT        0b0101011   // Rect コマンド
#define KNJLCD_CMD_GRAPH_BITMAP      0b01011100  // Bitmap コマンド
#define KNJLCD_CMD_GRAPH_FLIP        0b01011110  // Flip コマンド
#define KNJLCD_CMD_DIGITAL_WRITE     0b0110      // デジタル出力コマンド
#define KNJLCD_CMD_PIN_MODE          0b01110     // ピンモード設定コマンド
#define KNJLCD_CMD_PWM_WRITE         0b01111     // PWM 出力コマンド
#define KNJLCD_CMD_TEXT              0b1         // Text コマンド


#define KNJLCD_PORT_BACKLIGHT  7  // SG12832A の Backlight ON/OFF用


// Raspberry Pi に使用されている Broadcom BCM2835 の i2c には、clock stretching のバグがあります。
//     ⇒ 詳細は、http://www.advamation.com/knowhow/raspberrypi/rpi-i2c-bug.html を参照（英文です）

// i2c 通信は、例えば、マスターからスレーブへの書き込み(SlaveRecieve)の場合、
//  (1) i2c バスは解放されている
//  (2) マスターが start condition を発行し、i2c バスの制御権を掌握する
//  (3) マスターが、通信したいスレーブのアドレスと R/W bit を送信する
//  (4) 該当のスレーブが ACK か NACK を返送する
//  (5) NACK だった場合、(9) に行く
//  (6) ACK だった場合、マスターはデータ 8bit を送信する
//  (7) 該当のスレーブは送信されたデータ 8bit を受信し、ACK か NACK を返送する
//  (8) ACK なら、(6)～(7) を繰り返す
//  (9) NACK だった時や、送るべきデータがなくなった時は、マスターは stop condition を発行する
//  (10) i2c バスは解放される
// という感じで i2c 通信が行われますが、(4)や(7)の時、スレーブが ACK か NACK を返送する時
// スレーブ側の要請で clock stretching が発生する場合が有ります。
//（100％発生ではなく、低速応答のスレーブでのみ発生し、十分に高速なスレーブなら発生しない）

// マスターとなるべき Broadcom BCM2835 は、この clock stretching が起きた時の処理にバグがあるため、
// clock stretching が発生する i2c デバイスとは、上手く i2c 通信できない場合があります。
// 具体的には、上記(6)の時の、データ 8bit 送信での１ビット目のクロック幅が、仕様を満たさない状態になります。
// この為、スレーブ側では、数バイトは正常に受信するのに、次の数バイトは１ビットずつズレてしまう現象や、
// スレーブ側が 8bit 分を受信する前に stop condition が発行されて、スレーブ側が対応に困る、という現象が発生するようです。
// このバグは BCM2835 のハードウエア・レベルのバグですので、ソフト的に回避する方法は無いようです。

// 漢字 LCD シールド(π用)は、i2c 通信のスレーブとして AVR を使用していますが、
// やはり clock stretching が発生するらしく(オシロで波形を見たわけでないので断言できない)、
// 当初の i2c 通信方法では、盛大にコケてました。
// そこで対策として採ったのが、スレーブアドレスそのものでデータ送信する方法です。
// BCM2835 のバグは、上記(6)以降で問題になるので、それよりも前のスレーブアドレス部分だけは
// 安定して送信できているだろう、と考えたからです。
// 具体的には４つのスレーブアドレスを確保して、１つをパケット開始に、１つをパケット終了に、
// 残りの２つをデータ0/1用に使用する事にしました。

// ただし、この方法でも、マスターからスレーブへの送信はできますが、スレーブからマスターへの送信はできません。
// スレーブから送信されるデータ 8bit も clock stretching バグの影響を受けるからです。
// 唯一、スレーブアドレス直後に返送する ACK/NACK を使って 1bit だけスレーブのデータを送信する方法は可能です。
// 漢字 LCD シールド(π用)では、コマンド受付け可の時は ACK を、コマンド処理中で新規コマンド受付け不可の時は NACK を返す様にしています。
// (スレーブアドレスが４つあるので、各々の ACK/NACK に別々の意味を持たせれば 4bit のデータ送信も可能になりますが、そこまではサポートしませんでした)


// 漢字 LCD シールド(π用)管理構造体
struct knjlcdDataStruct
{
	int     fdStop;
	int     fdData1;
	int     fdData0;
	int     fdStart;
};

struct knjlcdDataStruct knjlcds [MAX_KNJLCDS];  // 最大４台の漢字LCDシールド(π用)を管理する：正直、４台も接続する事ないと思いますが、wiringPiのlcd.hに合わせておきます



// 漢字 LCD シールド(π用)がコマンド受付け可能になるまで待つ
// 戻り値が KNJLCD_ERR_NONE なら、コマンド受付け可能状態
// 注意：シールドの制御IC (ATmega328) の clock は約 8MHz なので、それなりに待ち時間を入れる必要があります。
static int waitWhileBusy (int fd)
{
	int res;
	unsigned int t = micros(); // wiringPiSetup() からの経過時間(μ秒)

	delayMicroseconds (10); // slave側の処理時間の確保

	while (1) { // 永久ループ
		res = wiringPiI2CWriteQuick (fd); // stop bit を送ってみる
		if (res == KNJLCD_ERR_NONE) {
			// ACK が返ってきた
			delayMicroseconds (10); // slave側の処理時間の確保
			return KNJLCD_ERR_NONE; // コマンド受付け可能状態
		}
//		printf("waitWhileBusy(): res = %02x\n", res);
		// 「まだだよ」返答：漢字 LCD シールドは以前のコマンドを処理中(SLA+W に対して NACK が返ってきた)か、そもそも、TWI 通信自体ができていない
		if ((micros() - t) > KNJLCD_CMD_TIMEOUT * 1000) { // 所定の時間が経過した (おそらく TWI 通信できていない)
			break; // 強制ループアウトさせる
		}
		delayMicroseconds (50); // 50μ秒待つ
	}
//	printf("waitWhileBusy(): TimeOut.\n");
	return KNJLCD_ERR_TIME_OUT; // タイムアウトした
}


// 漢字 LCD シールド(π用)に i2c 通信で bit を送信する
static int sendI2CBit (int fd)
{
	int res = wiringPiI2CWriteQuick (fd); // bit を送る

	delayMicroseconds (10); // slave側の処理時間の確保

	return res;
}

// len ビット分の bit 列を送信する
static int sendI2CBits (int fdIdx, unsigned long bits, int len)
{
	int res = -1;
	while (len) {
		--len;
		res = sendI2CBit ((bits & (1uL << len)) ? knjlcds[fdIdx].fdData1 : knjlcds[fdIdx].fdData0);
	}
	return res;
}


/*
 *********************************************************************************
 * User Callable code below here
 *********************************************************************************
 */



/*
 * knjlcdInit:
 *	Take a lot of parameters and initialise the LCD, and return a handle to
 *	that LCD, or -1 if any error.
   シールド上の DIPSW3 で設定している i2c アドレス(7bit format)を指定して下さい。
   建前上、file descriptor を返す、事になってますが、実際は knjlcds 配列のインデックス値(0～3)です。
   エラーの時は -1 を返します。
   正常に実行された場合は、画面クリア、カーソル位置 (0,0)、LCD 表示、カーソル非点滅、カーソル非表示、
   バックライト・ソフトOFF、汎用出力ピン All デジタル出力OFF になります。
 *********************************************************************************
 */

int knjlcdInit (int i2c_addr)
{
	static int initialised = 0;

	int i, res;
	int fdIdx = -1;

	if (initialised == 0) { // 初期化されていない
		initialised = 1;
		for (i = 0; i < MAX_KNJLCDS; ++i) {
			knjlcds[i].fdStop = -1;
		}
	}

	// 空いている漢字 LCD シールド(π用)管理構造体を探します。
	for (i = 0; i < MAX_KNJLCDS; ++i) {
		if (knjlcds[i].fdStop == -1) {
			fdIdx = i;
			break;
		}
	}

	if (fdIdx == -1) {
		// 既に4台確保している時はエラー
		return KNJLCD_ERR_FULL;
	}

	// 初期化
	// i2c アドレス連続４つを確保します。
	knjlcds[fdIdx].fdStop = wiringPiI2CSetup (i2c_addr + KNJLCD_I2C_STOP);
	if (knjlcds[fdIdx].fdStop == -1) {
		return KNJLCD_ERR_NO_DEV;
	}
	knjlcds[fdIdx].fdData1 = wiringPiI2CSetup (i2c_addr + KNJLCD_I2C_DATA1);
	if (knjlcds[fdIdx].fdData1 == -1) {
		return KNJLCD_ERR_NO_DEV;
	}
	knjlcds[fdIdx].fdData0 = wiringPiI2CSetup (i2c_addr + KNJLCD_I2C_DATA0);
	if (knjlcds[fdIdx].fdData0 == -1) {
		return KNJLCD_ERR_NO_DEV;
	}
	knjlcds[fdIdx].fdStart = wiringPiI2CSetup (i2c_addr + KNJLCD_I2C_START);
	if (knjlcds[fdIdx].fdStart == -1) {
		return KNJLCD_ERR_NO_DEV;
	}

	res = waitWhileBusy (knjlcds[fdIdx].fdStop);

	if (res == KNJLCD_ERR_NONE) {

		// 漢字 LCD シールドへ初期化コマンド(0000 0001)を送信
		sendI2CBit (knjlcds[fdIdx].fdStart);

		sendI2CBits (fdIdx, KNJLCD_CMD_INIT, 8);

		sendI2CBit (knjlcds[fdIdx].fdStop);

		// 続けて、画面クリアコマンドも実行します
		knjlcdClear (fdIdx);
	}

//	printf("knjlcdInit(): done.\n");

	return fdIdx;
}


/*
 * knjlcdHome:
 *	Home the cursor.
    カーソルを (0,0) の位置にリターンさせます。
 *********************************************************************************
 */

int knjlcdHome (int fdIdx)
{
	int res = waitWhileBusy (knjlcds[fdIdx].fdStop);

	if (res == KNJLCD_ERR_NONE) {

		// Return Home コマンド (0000 0010)
		sendI2CBit (knjlcds[fdIdx].fdStart);

		sendI2CBits (fdIdx, KNJLCD_CMD_RET_HOME, 8);

		sendI2CBit (knjlcds[fdIdx].fdStop);
	}

	return res;
}

/*
 * knjlcdClear:
 *	clear the screen.
    画面をクリアします。クリア処理に約60ms かかります。
    このコマンドそのものは、漢字LCDシールドに指示したら即リターンしますので、この次の命令で最大 60ms 待たされる事になります。
    なので、knjlcdClear 関数と次の命令の間に、60ms 以上かかる別の処理をやらせると良いでしょう。
 *********************************************************************************
 */

int knjlcdClear (int fdIdx)
{
	int res = waitWhileBusy (knjlcds[fdIdx].fdStop);

	if (res == KNJLCD_ERR_NONE) {

		// 画面クリアコマンド (0000 0011)
		sendI2CBit (knjlcds[fdIdx].fdStart);

		sendI2CBits (fdIdx, KNJLCD_CMD_CLEAR, 8);

		sendI2CBit (knjlcds[fdIdx].fdStop);
	}

	return res;
}


/*
 * knjlcdDisplay:
 *	Turn the display on/off
    液晶画面の表示を ON/OFF します。
    この命令では backlight の方は無変化です。
    knjlcdInit 関数を呼び出した直後は ON になっています。
 *********************************************************************************
 */

int knjlcdDisplay (int fdIdx, int state)
{
	int res = waitWhileBusy (knjlcds[fdIdx].fdStop);

	if (res == KNJLCD_ERR_NONE) {

		// 画面表示 ON/OFF コマンド (0001 010s)
		sendI2CBit (knjlcds[fdIdx].fdStart);

		sendI2CBits (fdIdx, KNJLCD_CMD_DISPLAY, 7);

		sendI2CBit (state ? knjlcds[fdIdx].fdData1 : knjlcds[fdIdx].fdData0); // ON/OFF bit

		sendI2CBit (knjlcds[fdIdx].fdStop);
	}

	return res;
}

/*
 * knjlcdCursor: 
 *	Turn the cursor on/off
    カーソル表示を ON/OFF します。
    knjlcdInit 関数を呼び出した直後は OFF になっています。
 *********************************************************************************
 */

int knjlcdCursor (int fdIdx, int state)
{
	int res = waitWhileBusy (knjlcds[fdIdx].fdStop);

	if (res == KNJLCD_ERR_NONE) {

		// カーソル表示 ON/OFF コマンド (0001 001s)
		sendI2CBit (knjlcds[fdIdx].fdStart);

		sendI2CBits (fdIdx, KNJLCD_CMD_CURSOR, 7);

		sendI2CBit (state ? knjlcds[fdIdx].fdData1 : knjlcds[fdIdx].fdData0); // ON/OFF bit

		sendI2CBit (knjlcds[fdIdx].fdStop);
	}

	return res;
}

/*
 * knjlcdCursorBlink:
 *	Turn the cursor blinking on/off
    カーソルの点滅を ON/OFF します。
    knjlcdInit 関数を呼び出した直後は OFF になっています。
 *********************************************************************************
 */

int knjlcdCursorBlink (int fdIdx, int state)
{
	int res = waitWhileBusy (knjlcds[fdIdx].fdStop);

	if (res == KNJLCD_ERR_NONE) {

		// カーソル点滅 ON/OFF コマンド (0001 000s)
		sendI2CBit (knjlcds[fdIdx].fdStart);

		sendI2CBits (fdIdx, KNJLCD_CMD_BLINK, 7);

		sendI2CBit (state ? knjlcds[fdIdx].fdData1 : knjlcds[fdIdx].fdData0); // ON/OFF bit

		sendI2CBit (knjlcds[fdIdx].fdStop);
	}

	return res;
}


/*
 * knjlcdPosition:
 *	Update the position of the cursor on the display.
 *	Ignore invalid locations.
    カーソルの位置を指定します。指定できるのは
    col は 0～15 の範囲
    row は 0～1 の範囲、です。
 *********************************************************************************
 */

int knjlcdPosition (int fdIdx, unsigned int col, unsigned int row)
{
	if (col >= KNJLCD_MAX_COLS) { // 範囲外
		return KNJLCD_ERR_OUT_OF_RANGE; // 引数が不正
	}
	if (row >= KNJLCD_MAX_ROWS) { // 範囲外
		return KNJLCD_ERR_OUT_OF_RANGE; // 引数が不正
	}

	int res = waitWhileBusy (knjlcds[fdIdx].fdStop);

	if (res == KNJLCD_ERR_NONE) {

		// カーソル Locate コマンド (001 y xxxxxx) : 10bit長
		sendI2CBit (knjlcds[fdIdx].fdStart);

		sendI2CBits (fdIdx, KNJLCD_CMD_LOCATE, 3);

		sendI2CBits (fdIdx, row, 1);  // y座標指定が先

		sendI2CBits (fdIdx, col, 6);  // x座標指定は後

		sendI2CBit (knjlcds[fdIdx].fdStop);
	}

	return res;
}


/*
 * knjlcdPutchar:
 *	Send a data byte to be displayed on the display. We implement a very
 *	simple terminal here - with line wrapping, but no scrolling. Yet.
    漢字 LCD シールド(π用)に文字コード１バイトを送信します。
    漢字 LCD シールド(π用)の方でマルチバイト処理しますので、１バイトずつ転送するだけで OK です
 *********************************************************************************
 */

int knjlcdPutchar (int fdIdx, char ch)
{
	int res = waitWhileBusy (knjlcds[fdIdx].fdStop);

	if (res == KNJLCD_ERR_NONE) {

//		printf("knjlcdPutchar(): ch = %02x\n", ch);

		// テキスト転送コマンド (1 cccccccc) : 9bit長
		sendI2CBit (knjlcds[fdIdx].fdStart);

		sendI2CBit (knjlcds[fdIdx].fdData1);  // Textコマンド

		sendI2CBits (fdIdx, ch, 8);

		sendI2CBit (knjlcds[fdIdx].fdStop);
	}

	return res;
}


/*
 * knjlcdPuts:
 *	Send a string to be displayed on the display
    文字列を漢字LCDシールドに送信、表示させます。
    自動的に改行しないので、改行させたい場合は'\n'を指定する必要があります。
    正常表示できた場合は、バイト数を返します。
 *********************************************************************************
 */

int knjlcdPuts (int fdIdx, const char * str)
{
	char ch;
	int res, len = 0;

	if (str == NULL) {
		return KNJLCD_ERR_INVALID_PARAM;
	}

	while (ch = *str++) {
		res = knjlcdPutchar (fdIdx, ch);
		if (res != KNJLCD_ERR_NONE) {
			return res;
		}
		len++;
	}
	return len;
}


/*
 * knjlcdPrintf:
 *	Printf to an LCD display
    printf の書式に従って、文字列を漢字LCDシールドに送信、表示させます。
    正常表示できた場合は、バイト数を返します。
 *********************************************************************************
 */

int knjlcdPrintf (int fdIdx, const char * fmt, ...)
{
	va_list argp;
	char buffer [1024];

	if (fmt == NULL) {
		return KNJLCD_ERR_INVALID_PARAM;
	}

	va_start  (argp, fmt);
	vsnprintf (buffer, 1023, fmt, argp);
	va_end    (argp);

	return knjlcdPuts (fdIdx, buffer);
}



// シールド上の DIPSW3 で、バックライト・ソフトON/OFF モードにしている場合のみ有効。
// バックライト強制ON モードにしている場合は、以下のコマンドに関係無く、バックライトON されます。
// knjlcdInit 関数を呼び出した直後は、バックライト・ソフトOFF になります。
int knjlcdBacklight (int fdIdx, int state)
{
	int res = waitWhileBusy (knjlcds[fdIdx].fdStop);

	if (res == KNJLCD_ERR_NONE) {

		// デジタル出力コマンド (0110 111s)
		sendI2CBit (knjlcds[fdIdx].fdStart);

		sendI2CBits (fdIdx, KNJLCD_CMD_DIGITAL_WRITE, 4);

		sendI2CBits (fdIdx, KNJLCD_PORT_BACKLIGHT, 3);

		sendI2CBit (state ? knjlcds[fdIdx].fdData1 : knjlcds[fdIdx].fdData0);  // ON/OFF bit

		sendI2CBit (knjlcds[fdIdx].fdStop);
	}

	return res;
}



// グラフィック描画
// 一旦、漢字 LCD コントローラ内の仮想画面(横128×縦32)に描いておいてから、flip 関数で一括して LCD へ転送する事で、初めて表示されます。
// 引数の x/x0/x1 は 0～127、y/y0/y1 は 0～31、color は 0 の時は消去、0 以外の時は描画、です。
// 注意１：文字を書くと、半角文字なら 16x8 単位、全角文字なら 16x16 単位で、その部分のグラフィックが消えます。つまり、文字とグラフィックは重ねられません。
// 注意２：激しく遅いです。過度の期待はしないように。


// 仮想画面に点を描画
int knjlcdPset (int fdIdx, int x, int y, int color)
{
	if ((x < 0) || (x >= KNJLCD_MAX_WIDTH)) {
		return KNJLCD_ERR_OUT_OF_RANGE;
	}
	if ((y < 0) || (y >= KNJLCD_MAX_HEIGHT)) {
		return KNJLCD_ERR_OUT_OF_RANGE;
	}

	int res = waitWhileBusy (knjlcds[fdIdx].fdStop);

	if (res == KNJLCD_ERR_NONE) {

		// 漢字 LCD シールドの仮想画面へ点描画 x, y, color
		// PSET コマンド (0101 000c xxxxxxx yyyyy) : 20bit長
		sendI2CBit (knjlcds[fdIdx].fdStart);

		sendI2CBits (fdIdx, KNJLCD_CMD_GRAPH_PSET, 7);

		sendI2CBit (color ? knjlcds[fdIdx].fdData1 : knjlcds[fdIdx].fdData0);  // 色bit：0(ドット無し)か1(ドット有り)

		sendI2CBits (fdIdx, x, 7);
		sendI2CBits (fdIdx, y, 5);

		sendI2CBit (knjlcds[fdIdx].fdStop);
	}

	return res;
}

// 仮想画面に線を描画
// 注意：開始位置(x0,y0)・終了位置(x1,y1)ともに、仮想画面の範囲内に存在していなければなりません。
int knjlcdLine (int fdIdx, int x0, int y0, int x1, int y1, int color)
{
	int res = waitWhileBusy (knjlcds[fdIdx].fdStop);

	if (res == KNJLCD_ERR_NONE) {

		if (x0 < 0) {
			x0 = 0;
		} else if (x0 >= KNJLCD_MAX_WIDTH) {
			x0 = KNJLCD_MAX_WIDTH - 1;
		}

		if (y0 < 0) {
			y0 = 0;
		} else if (y0 >= KNJLCD_MAX_HEIGHT) {
			y0 = KNJLCD_MAX_HEIGHT - 1;
		}

		if (x1 < 0) {
			x1 = 0;
		} else if (x1 >= KNJLCD_MAX_WIDTH) {
			x1 = KNJLCD_MAX_WIDTH - 1;
		}

		if (y1 < 0) {
			y1 = 0;
		} else if (y1 >= KNJLCD_MAX_HEIGHT) {
			y1 = KNJLCD_MAX_HEIGHT - 1;
		}

		// 漢字 LCD シールドの仮想画面へ線描画 x0, y0, x1, y1, color
		// LINE コマンド (0101 001c xxxxxxx yyyyy xxxxxxx yyyyy) : 32bit長
		sendI2CBit (knjlcds[fdIdx].fdStart);

		sendI2CBits (fdIdx, KNJLCD_CMD_GRAPH_LINE, 7);

		sendI2CBit (color ? knjlcds[fdIdx].fdData1 : knjlcds[fdIdx].fdData0);  // 色bit：0(ドット無し)か1(ドット有り)

		sendI2CBits (fdIdx, x0, 7);
		sendI2CBits (fdIdx, y0, 5);

		sendI2CBits (fdIdx, x1, 7);
		sendI2CBits (fdIdx, y1, 5);

		sendI2CBit (knjlcds[fdIdx].fdStop);
	}

	return res;
}

// 仮想画面に四角形(塗りつぶし)を描画
int knjlcdRect (int fdIdx, int x0, int y0, int x1, int y1, int color)
{
	int res = waitWhileBusy (knjlcds[fdIdx].fdStop);

	if (res == KNJLCD_ERR_NONE) {

		if (x0 < 0) {
			x0 = 0;
		} else if (x0 >= KNJLCD_MAX_WIDTH) {
			x0 = KNJLCD_MAX_WIDTH - 1;
		}

		if (y0 < 0) {
			y0 = 0;
		} else if (y0 >= KNJLCD_MAX_HEIGHT) {
			y0 = KNJLCD_MAX_HEIGHT - 1;
		}

		if (x1 < 0) {
			x1 = 0;
		} else if (x1 >= KNJLCD_MAX_WIDTH) {
			x1 = KNJLCD_MAX_WIDTH - 1;
		}

		if (y1 < 0) {
			y1 = 0;
		} else if (y1 >= KNJLCD_MAX_HEIGHT) {
			y1 = KNJLCD_MAX_HEIGHT - 1;
		}

		// 漢字 LCD シールドの仮想画面へ矩形描画 x0, y0, x1, y1, color
		// RECT コマンド (0101 010c xxxxxxx yyyyy xxxxxxx yyyyy) : 32bit長
		sendI2CBit (knjlcds[fdIdx].fdStart);

		sendI2CBits (fdIdx, KNJLCD_CMD_GRAPH_RECT, 7);

		sendI2CBit (color ? knjlcds[fdIdx].fdData1 : knjlcds[fdIdx].fdData0);  // 色bit：0(ドット無し)か1(ドット有り)

		sendI2CBits (fdIdx, x0, 7);
		sendI2CBits (fdIdx, y0, 5);

		sendI2CBits (fdIdx, x1, 7);
		sendI2CBits (fdIdx, y1, 5);

		sendI2CBit (knjlcds[fdIdx].fdStop);
	}

	return res;
}

// 仮想画面に Bitmap(モノクロ８画素単位)を１ライン分転送：
//    x は８画素単位(0,8,16,...,120)の指定。size は転送バイト数(最大16バイト)。画面右端を超える分の bitmap は無視される。
int knjlcdBitmap (int fdIdx, int x, int y, unsigned int size, const unsigned char * bitmap)
{
	int i, res;

	if ((x < 0) || (x >= KNJLCD_MAX_WIDTH)) {
		return KNJLCD_ERR_OUT_OF_RANGE;
	}
	if ((y < 0) || (y >= KNJLCD_MAX_HEIGHT)) {
		return KNJLCD_ERR_OUT_OF_RANGE;
	}

	if (size > 16) {
		size = 16; // 最大でも 16 byte : 漢字 LCD シールド側の表示幅(最大128画素)の制約にあわせる。
	}

	res = waitWhileBusy (knjlcds[fdIdx].fdStop);

	if (res == KNJLCD_ERR_NONE) {

		// 漢字 LCD シールドの仮想画面へ Bitmap 転送 x, y, bitmap, ...
		// BITMAP コマンド (0101 1100 xxxx yyyyy bbbbbbbb .... bbbbbbbb) : 25～145bit長
		sendI2CBit (knjlcds[fdIdx].fdStart);

		sendI2CBits (fdIdx, KNJLCD_CMD_GRAPH_BITMAP, 8);

		sendI2CBits (fdIdx, x >> 3, 4);  // 下位 3bit は無視
		sendI2CBits (fdIdx, y,      5);

		for (i = 0; i < size; i++) { // bitmap(8画素分) * size個
			sendI2CBits (fdIdx, bitmap[i], 8);
		}

		sendI2CBit (knjlcds[fdIdx].fdStop);
	}

	return res;
}

// 仮想画面から LCD のグラフィック面に転送：
// ここで初めて LCD に表示されるようになります。 
// ※ 処理に最大約 60ms かかります。領域が狭ければ狭いほど、処理が早く終わります。
int knjlcdFlip (int fdIdx, int x0, int y0, int x1, int y1)
{
	int res = waitWhileBusy (knjlcds[fdIdx].fdStop);

	if (res == KNJLCD_ERR_NONE) {

		if (x0 < 0) {
			x0 = 0;
		} else if (x0 >= KNJLCD_MAX_WIDTH) {
			x0 = KNJLCD_MAX_WIDTH - 1;
		}

		if (y0 < 0) {
			y0 = 0;
		} else if (y0 >= KNJLCD_MAX_HEIGHT) {
			y0 = KNJLCD_MAX_HEIGHT - 1;
		}

		if (x1 < 0) {
			x1 = 0;
		} else if (x1 >= KNJLCD_MAX_WIDTH) {
			x1 = KNJLCD_MAX_WIDTH - 1;
		}

		if (y1 < 0) {
			y1 = 0;
		} else if (y1 >= KNJLCD_MAX_HEIGHT) {
			y1 = KNJLCD_MAX_HEIGHT - 1;
		}

		// 漢字 LCD シールドの仮想画面を表示面に転送 x0, y0, x1, y1
		// FLIP コマンド (0101 1110 xxxx yyyyy xxxx yyyyy) : 26bit長
		sendI2CBit (knjlcds[fdIdx].fdStart);

		sendI2CBits (fdIdx, KNJLCD_CMD_GRAPH_FLIP, 8);

		sendI2CBits (fdIdx, x0 >> 3, 4);  // 下位 3bit は無視
		sendI2CBits (fdIdx, y0,      5);

		sendI2CBits (fdIdx, x1 >> 3, 4);  // 下位 3bit は無視
		sendI2CBits (fdIdx, y1,      5);

		sendI2CBit (knjlcds[fdIdx].fdStop);
	}

	return res;
}



// 汎用出力関係
// 注意：デジタル出力も PWM 出力も、共に +5V 出力になります。+3.3V ではありませんので、注意して下さい。

// ※ 当初、姉妹製品の"カナLCDシールド(π用)"にて、PWM 機能が 2ch しかない ATtiny48 を使用する予定だった為、PWM0 と PWM1 の２つのみになる予定でした（OUT0 はデジタル出力のみだった）。
//    しかし、諸事情で ATmega48A に変更になった時に、OUT0 を PWM2 に名称変更するべきだったのですが、思いっきり見落として、基板を作ってしまいました。
//    (本製品の方は、最初から ATmega328 を使用するつもりだったので、この事に直ぐに気がついてもよかったハズなんですが…)
//    分かりやすさから言えば、PWM0/PWM1/PWM2/OUT3 という風にした方が絶対に良かったのですが、基板作り直しにお金が掛かるので、申し訳有りませんが、'現状優先'とさせて下さい。


// 指定の汎用出力ピン(PWM0/PWM1/OUT0)を OUTPUTモードか PWM_OUTPUT モードに設定します。
// knjlcdInit 関数を呼び出した直後は、３点とも OUTPUTモードになります。
int knjlcdPinMode (int fdIdx, unsigned int pin, int mode)
{
	if (pin >= KNJLCD_MAX_PINS) {
		return KNJLCD_ERR_OUT_OF_RANGE;
	}

	int res = waitWhileBusy (knjlcds[fdIdx].fdStop);

	if (res == KNJLCD_ERR_NONE) {

		// ピンモード設定コマンド (0111 0ppm)
		sendI2CBit (knjlcds[fdIdx].fdStart);

		sendI2CBits (fdIdx, KNJLCD_CMD_PIN_MODE, 5);

		sendI2CBits (fdIdx, pin, 2);  // 0～2のみ。3は何も影響しない。

		sendI2CBit ((mode == PWM_OUTPUT) ? knjlcds[fdIdx].fdData1 : knjlcds[fdIdx].fdData0);  // mode bit

		sendI2CBit (knjlcds[fdIdx].fdStop);
	}

	return res;
}


// 指定の汎用出力ピン(PWM0/PWM1/OUT0/OUT1)をデジタル ON/OFF させます。
// knjlcdInit 関数を呼び出した直後は、４点とも OFF になります。
// 注意：+5V 仕様です。+3.3V ではありません。
int knjlcdDigitalWrite (int fdIdx, unsigned int pin, int state)
{
	if (pin >= KNJLCD_MAX_PINS) {
		return KNJLCD_ERR_OUT_OF_RANGE;
	}

	int res = waitWhileBusy (knjlcds[fdIdx].fdStop);

	if (res == KNJLCD_ERR_NONE) {

		// デジタル出力コマンド (0110 ppps)
		sendI2CBit (knjlcds[fdIdx].fdStart);

		sendI2CBits (fdIdx, KNJLCD_CMD_DIGITAL_WRITE, 4);

		sendI2CBits (fdIdx, pin, 3);  // 0～3のみ。4～6は何も影響しない。7はBacklightで使用。

		sendI2CBit (state ? knjlcds[fdIdx].fdData1 : knjlcds[fdIdx].fdData0);  // ON/OFF bit

		sendI2CBit (knjlcds[fdIdx].fdStop);
	}

	return res;
}


// 指定の汎用出力ピン(PWM0/PWM1/OUT0)から PWM 出力させます。
// knjlcdPinMode 関数で PWM_OUTPUTモードに設定している必要があります。
// 指定できる PWM 値は 0～255 の範囲です。
// 注意：+5V 仕様です。+3.3V ではありません。
int knjlcdPwmWrite (int fdIdx, unsigned int pin, unsigned int val)
{
	if (pin >= KNJLCD_MAX_PWMS) {
		return KNJLCD_ERR_OUT_OF_RANGE;
	}

	int res = waitWhileBusy (knjlcds[fdIdx].fdStop);

	if (res == KNJLCD_ERR_NONE) {

		// PWM出力コマンド (0111 1pp vvvvvvvv) : 15bit長
		sendI2CBit (knjlcds[fdIdx].fdStart);

		sendI2CBits (fdIdx, KNJLCD_CMD_PWM_WRITE, 5);

		sendI2CBits (fdIdx, pin, 2);  // 0～2のみ。3は何も影響しない。

		sendI2CBits (fdIdx, val, 8);  // PWM 値：0～255

		sendI2CBit (knjlcds[fdIdx].fdStop);
	}

	return res;
}

