/*
 * knjlcd.h:
 *	Text-based Kanji LCD shield for π's driver.
 *	This is designed to drive the I2C interface LCD drivers
 *
 * Copyright (c) 2013 I.Kato @ BS21 Lab
 *
 ***********************************************************************
 * This file is NOT a part of wiringPi, but You have to add this to wiringPi for using.
 *
 *    knjlcd is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU Lesser General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    knjlcd is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU Lesser General Public License for more details.
 *
 *    You should have received a copy of the GNU Lesser General Public License
 *    along with wiringPi.  If not, see <http://www.gnu.org/licenses/>.
 ***********************************************************************
 */


// 漢字 LCD シールド(π用)

// RaspberryPi から i2c通信で SG12832A を制御します。
// 使用する i2c アドレスは、0x0C～0x0F, 0x1C～0x1F, 0x2C～0x2F, 0x3C～0x3F の４つから選択できます。
// 必ず、連続４アドレス（例えば、0x3C,0x3D,0x3E,0x3F）をごっそりと占有しますので、注意して下さい。

// 表示できる文字は、JIS 第一水準字＋α、のみです。
// 使用できる符号化方式は UTF-8 のみです。他の符号化方式には変更できません。

// SG12832A 表示機能以外に、汎用出力ピン４点があります。ただし +5V 出力ですので、注意が必要です。
// 汎用出力ピン４点の内、３点は PWM 出力(8bit長)も可能です。こちらも +5V 出力になります。



// 最大４台の漢字 LCD シールド(π用)を管理します。
// 正直、４台も接続する事ないと思いますが、wiringPi の lcd.h に合わせておきます。
// 当然ですが、各々の i2c アドレスは重複しないように設定する必要が有ります。
#define MAX_KNJLCDS   4


// 漢字 LCD シールド(π用)の最大桁数・行数
#define KNJLCD_MAX_COLS    16  // 最大桁数(半角の時)
#define KNJLCD_MAX_ROWS     2  // 最大行数
#define KNJLCD_MAX_WIDTH  128  // 最大グラフィック表示幅
#define KNJLCD_MAX_HEIGHT  32  // 最大グラフィック表示高さ


#define KNJLCD_STATE_OFF   0
#define KNJLCD_STATE_ON    1


// 汎用出力ピン：４点。この内３点は PWM 出力も可能。
#define KNJLCD_PIN_PWM0   0  // シルク'PWM0' ※デジタルON/OFF出力も可能
#define KNJLCD_PIN_PWM1   1  // シルク'PWM1' ※デジタルON/OFF出力も可能
#define KNJLCD_PIN_OUT0   2  // シルク'OUT0' ※PWM出力も可能
#define KNJLCD_PIN_OUT1   3  // シルク'OUT1' ※デジタルON/OFF出力のみ。PWM出力はできない

#define KNJLCD_MAX_PINS   4  // デジタル出力できる最大ピン数
#define KNJLCD_MAX_PWMS   3  // PWM出力できる最大ピン数


// 漢字 LCD ライブラリ(π用)のエラーコード
#define KNJLCD_ERR_NONE             0    // 正常
#define KNJLCD_ERR_TIME_OUT       (-15)  // 漢字 LCD シールドとの通信がタイムアウトした
#define KNJLCD_ERR_NO_DEV         (-12)  // 指定した i2c アドレスの漢字 LCD シールドは無かった
#define KNJLCD_ERR_FULL           (-11)  // ４台以上を確保しようとした
#define KNJLCD_ERR_INSUF_MEM      (-10)  // メモリが足らない
#define KNJLCD_ERR_OUT_OF_RANGE   (-6)   // 指定できる範囲から逸脱している
#define KNJLCD_ERR_INVALID_PARAM  (-5)   // 引数が不正
#define KNJLCD_ERR_NOT_INIT       (-3)   // 初期化されていない
#define KNJLCD_ERR_UNKNOWN        (-1)   // 原因不明のエラーが発生した


#ifdef __cplusplus
extern "C" {
#endif


// 漢字 LCD シールド(π用)を初期化する。
// シールド上の DIPSW3 で設定している i2c アドレス(7bit format)を指定して下さい。
// 正常に実行された場合は、画面クリア、カーソル位置 (0,0)、LCD 表示、カーソル非点滅、カーソル非表示、バックライト・ソフトOFF、汎用出力ピン All OFF になります
extern int knjlcdInit (int i2c_addr);  // file descriptor を返します。エラーの時は -1


// wiringPi の lcd.h 互換関数

extern int knjlcdHome        (int fd);  // 画面はそのままで、カーソル位置だけ (0,0) にする
extern int knjlcdClear       (int fd);  // 画面をクリアし、カーソル位置を (0,0) にする ※処理に約 60ms かかります
extern int knjlcdDisplay     (int fd, int state);  // LCD を表示させる/させない
extern int knjlcdCursor      (int fd, int state);  // カーソルを表示させる/させない ※半角/全角に関係なく、常に全角文字単位で表示されます。
extern int knjlcdCursorBlink (int fd, int state);  // カーソル位置の文字を点滅させる/させない ※半角/全角に関係なく、常に全角文字単位で点滅されます。
extern int knjlcdPosition    (int fd, unsigned int col, unsigned int row);  // カーソル位置を設定: col=0～15, row=0～1 の範囲で指定 ※全角文字は偶数colのみ指定可能です
//extern int knjlcdCharDef   (int fd, unsigned int index, unsigned char font_data [8]);  // 有りません。代わりに knjlcdBitmap を使ってください。
extern int knjlcdPutchar     (int fd, char ch);  // 文字コードを１バイト転送する
extern int knjlcdPuts        (int fd, const char * str);  // 文字列を転送する。自動改行はしないので'\n'を指定する必要あり。
extern int knjlcdPrintf      (int fd, const char * fmt, ...);  // printf


// 漢字 LCD シールド(π用)専用の関数

//extern int knjlcdEntryModeSet (int fd, unsigned int ems);    // 有りません。
//extern int knjlcdShift        (int fd, unsigned int shift);  // 有りません。

// シールド上の DIPSW3 で、バックライト・ソフトON/OFF モードにしている場合のみ有効。
// バックライト強制ON モードにしている場合は、以下のコマンドに関係無く、バックライトON される
extern int knjlcdBacklight (int fd, int state); // LCD バックライトを ON/OFF させる


// グラフィック描画
// 一旦、漢字 LCD コントローラ内の仮想画面(横128×縦32)に描いておいてから、flip 関数で一括して LCD へ転送する事で、初めて表示されます。
// 引数の x/x0/x1 は 0～127、y/y0/y1 は 0～31、color は 0 の時は消去、0 以外の時は描画、です。
// 注意１：文字を書くと、半角文字なら 16x8 単位、全角文字なら 16x16 単位で、その部分のグラフィックが消えます。つまり、文字とグラフィックは重ねられません。
// 注意２：激しく遅いです。過度の期待はしないように。

extern int knjlcdPset   (int fd, int x,  int y,  int color); // 仮想画面に点を描画
extern int knjlcdLine   (int fd, int x0, int y0, int x1, int y1, int color); // 仮想画面に線を描画
extern int knjlcdRect   (int fd, int x0, int y0, int x1, int y1, int color); // 仮想画面に四角形(塗りつぶし)を描画
extern int knjlcdBitmap (int fd, int x,  int y,  unsigned int size, const unsigned char * bitmap); // 仮想画面に Bitmap(モノクロ８画素単位)を１ライン分転送：x は８画素単位(0,8,16,...,120)の指定。size は転送バイト数。画面右端を超える分の bitmap は無視される。
extern int knjlcdFlip   (int fd, int x0, int y0, int x1, int y1); // 仮想画面から LCD のグラフィック面に転送：ここで初めて LCD に表示される ※処理に最大約 60ms かかります


// 汎用出力ピン
extern int knjlcdPinMode      (int fd, unsigned int pin, int mode);          // 指定の汎用出力ピン(PWM0/PWM1/OUT0)を OUTPUTモードか PWM_OUTPUT モードに設定する
extern int knjlcdDigitalWrite (int fd, unsigned int pin, int state);         // 指定の汎用出力ピン(PWM0/PWM1/OUT0/OUT1)をデジタル ON/OFF させる
extern int knjlcdPwmWrite     (int fd, unsigned int pin, unsigned int val);  // 指定の汎用出力ピン(PWM0/PWM1/OUT0)から PWM 出力させる。PWM_OUTPUTモードになっている必要あり。指定できる値は 0～255 の範囲。


#ifdef __cplusplus
}
#endif

